"""Module with methods to propagate the uncertainties through the raking process"""

import numpy as np

from scipy.linalg import lu_factor, lu_solve


def compute_covariance(
    Dphi_y: np.ndarray,
    Dphi_s: np.ndarray,
    sigma_yy: np.ndarray,
    sigma_ss: np.ndarray,
    sigma_ys: np.ndarray,
) -> np.ndarray:
    """Compute the covariance matrix of the raked values.

    The covariance matrix of the raked values is phi' Sigma phi'T
    where phi' is the matrix of the partial derivatives of the raked values beta
    with respect to the observations y and margins s.

    Parameters
    ----------
    dPhi_y : np.ndarray
        Derivatives with respect to the observations
    Dphi_s : np.ndarray
        Derivatives with respect to the margins
    sigma_yy : np.ndarray
        Covariance matrix of the observations
    sigma_ss : np.ndarray
        Covariance matrix of the margins
    sigma_ys : np.ndarray
        Covariance matrix of the observations and margins

    Returns
    -------
    covariance : np.ndarray
        Covariance matrix of the raked values
    """
    assert isinstance(Dphi_y, np.ndarray), (
        "The derivatives matrix with respect to the observations should be a Numpy array."
    )
    assert len(Dphi_y.shape) == 2, (
        "The derivatives matrix with respect to the observations should be a 2D Numpy array."
    )
    assert np.shape(Dphi_y)[0] == np.shape(Dphi_y)[1], (
        "The derivatives matrix with respect to the observations should be a square matrix."
    )
    assert isinstance(Dphi_s, np.ndarray), (
        "The derivatives matrix with respect to the margins should be a Numpy array."
    )
    assert len(Dphi_s.shape) == 2, (
        "The derivatives matrix with respect to the margins should be a 2D Numpy array."
    )
    assert isinstance(sigma_yy, np.ndarray), (
        "The covariance matrix of the observations should be a Numpy array."
    )
    assert len(sigma_yy.shape) == 2, (
        "The covariance matrix of the observations should be a 2D Numpy array."
    )
    assert np.shape(sigma_yy)[0] == np.shape(sigma_yy)[1], (
        "The covariance matrix of the observations should be a square matrix."
    )
    assert isinstance(sigma_ss, np.ndarray), (
        "The covariance matrix of the margins should be a Numpy array."
    )
    assert len(sigma_ss.shape) == 2, (
        "The covariance matrix of the margins should be a 2D Numpy array."
    )
    assert np.shape(sigma_ss)[0] == np.shape(sigma_ss)[1], (
        "The covariance matrix of the margins should be a square matrix."
    )
    assert isinstance(sigma_ys, np.ndarray), (
        "The covariance matrix of the observations and margins should be a Numpy array."
    )
    assert len(sigma_ys.shape) == 2, (
        "The covariance matrix of the observations and margins should be a 2D Numpy array."
    )
    assert np.shape(Dphi_y)[0] == np.shape(Dphi_s)[0], (
        "The derivative matrices with respect to the observations and the margins should have the same number of rows."
    )
    assert np.shape(Dphi_y)[0] == np.shape(sigma_yy)[0], (
        "The derivative matrix with respect to the observations and the covariance matrix of the observations should have the same size."
    )
    assert np.shape(Dphi_s)[1] == np.shape(sigma_ss)[1], (
        "The derivative matrix with respect to the margins and the covariance matrix of the margins should have the same number of columns."
    )
    assert np.shape(sigma_ys)[0] == np.shape(sigma_yy)[0], (
        "The covariance matrix of observations and margins should have the same number of rows as the covariance matrix of the observations."
    )
    assert np.shape(sigma_ys)[1] == np.shape(sigma_ss)[1], (
        "The covariance matrix of observations and margins should have the same number of columns as the covariance matrix of the margins."
    )

    Dphi = np.concatenate((Dphi_y, Dphi_s), axis=1)
    sigma = np.concatenate(
        (
            np.concatenate((sigma_yy, sigma_ys), axis=1),
            np.concatenate((np.transpose(sigma_ys), sigma_ss), axis=1),
        ),
        axis=0,
    )
    covariance = np.matmul(Dphi, np.matmul(sigma, np.transpose(Dphi)))
    return covariance


def compute_gradient(
    beta_0: np.ndarray,
    lambda_0: np.ndarray,
    y: np.ndarray,
    A: np.ndarray,
    method: str,
    alpha: float = 1,
    l: np.ndarray = None,
    h: np.ndarray = None,
    q: np.ndarray = None,
) -> tuple[np.ndarray, np.ndarray]:
    """Compute the gradient dbeta/dy and dbeta/ds.

    The covariance matrix of the raked values is phi' Sigma phi'T
    where phi' is the matrix of the partial derivatives of the raked values beta
    with respect to the observations y and margins s. This function computes phi'

    Parameters
    ----------
    beta_0 : np.ndarray
        Vector of raked values
    lambda_0 : np.ndarray
        Corresponding dual
    y : np.ndarray
        Vector of observations
    A : np.ndarray
        Constraints matrix (output of a function from the compute_constraints module)
    method : string
        Raking method (one of chi2, entropic, general, logit)
    alpha : float
        Parameter of the distance function, alpha=1 is the chi2 distance, alpha=0 is the entropic distance
    l : np.ndarray
        Lower bounds for the observations
    h : np.ndarray
        Upper bounds for the observations
    q :  np.ndarray
        Vector of weights (default to all 1)

    Returns
    -------
    Dphi_y : np.ndarray
        Derivatives with respect to the observations
    Dphi_s: np.ndarray
        Derivatives with respect to the margins
    """
    assert isinstance(beta_0, np.ndarray), (
        "The vector of raked values should be a Numpy array."
    )
    assert len(beta_0.shape) == 1, (
        "The vector of raked values should be a 1D Numpy array."
    )
    assert isinstance(lambda_0, np.ndarray), (
        "The dual vector should be a Numpy array."
    )
    assert len(lambda_0.shape) == 1, (
        "The vdual vector should be a 1D Numpy array."
    )
    assert isinstance(y, np.ndarray), (
        "The vector of observations should be a Numpy array."
    )
    assert len(y.shape) == 1, (
        "The vector of observations should be a 1D Numpy array."
    )
    assert len(y) == len(beta_0), (
        "The vectors of observations and raked values should have the same length."
    )
    assert isinstance(A, np.ndarray), (
        "The constraint matrix should be a Numpy array."
    )
    assert len(A.shape) == 2, (
        "The constraints matrix should be a 2D Numpy array."
    )
    assert np.shape(A)[0] == len(lambda_0), (
        "The number of linear constraints should be equal to the length of the dual vector."
    )
    assert np.shape(A)[1] == len(y), (
        "The number of coefficients for the linear constraints should be equal to the number of observations."
    )
    assert method in [
        "chi2",
        "entropic",
        "general",
        "logit",
    ], 'The raking method must be "chi2", "entropic", "general", or "logit".'
    if method == "general":
        assert isinstance(alpha, (int, float)), (
            "The parameter of the distance function should be an integer or a float."
        )
    if method == "logit":
        if l is None:
            l = np.zeros(len(y))
        assert isinstance(l, np.ndarray), (
            "The vector of lower bounds should be a Numpy array."
        )
        assert len(l.shape) == 1, (
            "The vector of lower bounds should be a 1D Numpy array."
        )
        assert len(y) == len(l), (
            "Observations and lower bounds vectors should have the same length."
        )
        assert np.all(l >= 0.0), "The lower bounds must be positive."
        assert np.all(l <= y), (
            "The observations must be superior or equal to the corresponding lower bounds."
        )
        if h is None:
            h = np.ones(len(y))
        assert isinstance(h, np.ndarray), (
            "The vector of upper bounds should be a Numpy array."
        )
        assert len(h.shape) == 1, (
            "The vector of upper bounds should be a 1D Numpy array."
        )
        assert len(y) == len(h), (
            "Observations and upper bounds vectors should have the same length."
        )
        assert np.all(h > 0.0), "The upper bounds must be strictly positive."
        assert np.all(h >= y), (
            "The observations must be inferior or equal to the correspondings upper bounds."
        )
        assert np.all(l < h), (
            "The lower bounds must be stricty inferior to the correspondings upper bounds."
        )
    if q is not None:
        assert isinstance(q, np.ndarray), (
            "The vector of weights should be a Numpy array."
        )
        assert len(q.shape) == 1, (
            "The vector of weights should be a 1D Numpy array."
        )
        assert len(y) == len(q), (
            "Observations and weights vectors should have the same length."
        )

    if q is None:
        q = np.ones(len(y))

    # Partial derivatives of the distance function with respect to raked values and observations
    if method == "chi2":
        DF1_beta_diag = np.zeros(len(beta_0))
        DF1_beta_diag[y != 0] = 1.0 / (q[y != 0] * y[y != 0])
        DF1_beta_diag[y == 0] = 0.0
        DF1_beta = np.diag(DF1_beta_diag)
        DF1_y_diag = np.zeros(len(y))
        DF1_y_diag[y != 0] = -beta_0[y != 0] / (
            q[y != 0] * np.square(y[y != 0])
        )
        DF1_y_diag[y == 0] = 0.0
        DF1_y = np.diag(DF1_y_diag)
    elif method == "entropic":
        DF1_beta_diag = np.zeros(len(beta_0))
        DF1_beta_diag[beta_0 != 0] = 1.0 / (
            q[beta_0 != 0] * beta_0[beta_0 != 0]
        )
        DF1_beta_diag[beta_0 == 0] = 0.0
        DF1_beta = np.diag(DF1_beta_diag)
        DF1_y_diag = np.zeros(len(y))
        DF1_y_diag[y != 0] = -1.0 / (q[y != 0] * y[y != 0])
        DF1_y_diag[y == 0] = 0.0
        DF1_y = np.diag(DF1_y_diag)
    elif method == "general":
        DF1_beta_diag = np.zeros(len(beta_0))
        DF1_beta_diag[(y != 0) & (beta_0 != 0)] = np.power(
            beta_0[(y != 0) & (beta_0 != 0)], alpha - 1.0
        ) / (
            q[(y != 0) & (beta_0 != 0)]
            * np.power(y[(y != 0) & (beta_0 != 0)], alpha)
        )
        DF1_beta_diag[(y == 0) | (beta_0 == 0)] = 0.0
        DF1_beta = np.diag(DF1_beta_diag)
        DF1_y_diag = np.zeros(len(y))
        DF1_y_diag[(y != 0) & (beta_0 != 0)] = -np.power(
            beta_0[(y != 0) & (beta_0 != 0)], alpha
        ) / (
            q[(y != 0) & (beta_0 != 0)]
            * np.power(y[(y != 0) & (beta_0 != 0)], alpha + 1.0)
        )
        DF1_y_diag[(y == 0) | (beta_0 == 0)] = 0.0
        DF1_y = np.diag(DF1_y_diag)
    elif method == "logit":
        DF1_beta_diag = np.zeros(len(beta_0))
        DF1_beta_diag[(beta_0 != l) & (beta_0 != h)] = 1.0 / (
            beta_0[(beta_0 != l) & (beta_0 != h)]
            - l[(beta_0 != l) & (beta_0 != h)]
        ) + 1.0 / (
            h[(beta_0 != l) & (beta_0 != h)]
            - beta_0[(beta_0 != l) & (beta_0 != h)]
        )
        DF1_beta_diag[(beta_0 == l) | (beta_0 == h)] = 0.0
        DF1_beta = np.diag(DF1_beta_diag)
        DF1_y_diag = np.zeros(len(y))
        DF1_y_diag[(y != l) & (y != h)] = -1.0 / (
            y[(y != l) & (y != h)] - l[(y != l) & (y != h)]
        ) - 1.0 / (h[(y != l) & (y != h)] - y[(y != l) & (y != h)])
        DF1_y_diag[(y == l) | (y == h)] = 0.0
        DF1_y = np.diag(DF1_y_diag)

    # Gradient with respect to beta and lambda
    DF1_lambda = np.transpose(np.copy(A))
    DF2_beta = np.copy(A)
    DF2_lambda = np.zeros((np.shape(A)[0], np.shape(A)[0]))
    DF_beta_lambda = np.concatenate(
        (
            np.concatenate((DF1_beta, DF1_lambda), axis=1),
            np.concatenate((DF2_beta, DF2_lambda), axis=1),
        ),
        axis=0,
    )

    # Gradient with respect to y and s
    DF1_s = np.zeros((np.shape(A)[1], np.shape(A)[0]))
    DF2_y = np.zeros((np.shape(A)[0], np.shape(A)[1]))
    DF2_s = -np.identity(np.shape(A)[0])
    DF_y_s = np.concatenate(
        (
            np.concatenate((DF1_y, DF1_s), axis=1),
            np.concatenate((DF2_y, DF2_s), axis=1),
        ),
        axis=0,
    )

    # Solve system DF_beta_lambda Dphi_y_s = - DF_y_s
    Dphi_y_s = np.zeros_like(DF_y_s)
    lu, piv = lu_factor(DF_beta_lambda)
    for i in range(0, np.shape(DF_y_s)[1]):
        Dphi_y_s[:, i] = -lu_solve((lu, piv), DF_y_s[:, i])

    # Return gradient of beta and lambda with respect to y and s
    Dphi_y = Dphi_y_s[0 : np.shape(A)[1], 0 : np.shape(A)[1]]
    Dphi_s = Dphi_y_s[
        0 : np.shape(A)[1], np.shape(A)[1] : (np.shape(A)[0] + np.shape(A)[1])
    ]
    return (Dphi_y, Dphi_s)



import pytest
import numpy as np
from raking.compute_constraints import (
    constraints_1D,
    constraints_2D,
    constraints_3D,
    constraints_USHD,
    constraints_USHD_lower,
)


def test_constraints_1D():
    # Generate balanced vector
    I = 3
    rng = np.random.default_rng(0)
    beta = rng.uniform(low=2.0, high=3.0, size=I)
    s = np.sum(beta)
    # Generate the constraints
    (A, s) = constraints_1D(s, I)
    # Verify that the constraint A beta = s is respected
    assert np.allclose(np.matmul(A, beta), s), (
        "For the constraints_1D function, the constraint A beta = s is not respected."
    )
    # Verify that the matrix A has rank 1
    assert np.linalg.matrix_rank(A) == 1, (
        "The constraint matrix should have rank 1."
    )


def test_constraints_2D():
    # Generate balanced matrix
    I = 3
    J = 5
    rng = np.random.default_rng(0)
    beta = rng.uniform(low=2.0, high=3.0, size=(I, J))
    s1 = np.sum(beta, axis=0)
    s2 = np.sum(beta, axis=1)
    beta = beta.flatten(order="F")
    # Generate the constraints
    (A, s) = constraints_2D(s1, s2, I, J)
    # Verify that the constraint A beta = s is respected
    assert np.allclose(np.matmul(A, beta), s), (
        "For the constraints_2D function, the constraint A beta = s is not respected."
    )
    # Verify that the matrix A has rank I + J - 1
    assert np.linalg.matrix_rank(A) == I + J - 1, (
        "The constraint matrix should have rank {}.".format(I + J - 1)
    )


def test_constraints_3D():
    # Generate balanced matrix
    I = 3
    J = 4
    K = 5
    rng = np.random.default_rng(0)
    beta = rng.uniform(low=2.0, high=3.0, size=(I, J, K))
    s1 = np.sum(beta, axis=0)
    s2 = np.sum(beta, axis=1)
    s3 = np.sum(beta, axis=2)
    beta = beta.flatten(order="F")
    # Generate the constraints
    (A, s) = constraints_3D(s1, s2, s3, I, J, K)
    # Verify that the constraint A beta = s is respected
    assert np.allclose(np.matmul(A, beta), s), (
        "For the constraints_3D function, the constraint A beta = s is not respected."
    )
    # Verify that the matrix A has rank I * J + I * K + J * K - I - J - K + 1
    assert np.linalg.matrix_rank(A) == I * J + I * K + J * K - I - J - K + 1, (
        "The constraint matrix should have rank {}.".format(
            I * J + I * K + J * K - I - J - K + 1
        )
    )


def test_constraints_USHD():
    # Generate balanced array
    I = 3
    J = 4
    K = 5
    rng = np.random.default_rng(0)
    beta_ijk = rng.uniform(low=2.0, high=3.0, size=(I, J, K))
    beta_00k = np.sum(beta_ijk, axis=(0, 1))
    beta_i0k = np.sum(beta_ijk, axis=1)
    beta_0jk = np.sum(beta_ijk, axis=0)
    beta1 = np.concatenate(
        (beta_00k.reshape((1, 1, K)), beta_i0k.reshape(I, 1, K)), axis=0
    )
    beta2 = np.concatenate((beta_0jk.reshape((1, J, K)), beta_ijk), axis=0)
    beta = np.concatenate((beta1, beta2), axis=1)
    beta = beta.flatten("F")
    beta_i = np.sum(beta_ijk, axis=(1, 2))
    beta_0 = np.sum(beta_i)
    s_cause = np.array([beta_0] + beta_i.tolist())
    # Generate the constraints
    (A, s) = constraints_USHD(s_cause, I, J, K)
    # Verify that the constraint A beta = s is respected
    assert np.allclose(np.matmul(A, beta), s), (
        "For the constraints_USHD function, the constraint A beta = s is not respected."
    )
    # Verify that the matrix A has rank I + 2 * K + J * K + (I - 1) * K
    assert np.linalg.matrix_rank(A) == I + 2 * K + J * K + (I - 1) * K, (
        "The constraint matrix should have rank {}.".format(
            I + 2 * K + J * K + (I - 1) * K
        )
    )


def test_constraints_USHD_lower():
    # Generate balanced array
    I = 3
    J = 4
    K = 5
    rng = np.random.default_rng(0)
    beta_ijk = rng.uniform(low=2.0, high=3.0, size=(I, J, K))
    s_cause = np.sum(beta_ijk, axis=(1, 2))
    s_county = np.sum(beta_ijk, axis=(0, 1))
    s_all_causes = np.sum(beta_ijk, axis=0)
    beta_i0k = np.sum(beta_ijk, axis=1)
    beta = np.concatenate((beta_i0k.reshape((I, 1, K)), beta_ijk), axis=1)
    beta = beta.flatten("F")
    # Generate the constraints
    (A, s) = constraints_USHD_lower(s_cause, s_county, s_all_causes, I, J, K)
    # Verify that the constraint A beta = s is respected
    assert np.allclose(np.matmul(A, beta), s), (
        "For the constraints_USHD_lower function, the constraint A beta = s is not respected."
    )
    # Verify that the matrix A has rank I + K + I * K + J * K - K - 1
    assert np.linalg.matrix_rank(A) == I + K + I * K + J * K - K - 1, (
        "The constraint matrix should have rank {}.".format(
            I + K + I * K + J * K - K - 1
        )
    )



import pytest
import numpy as np
from raking.run_raking import run_raking


def test_run_raking_1D(example_1D):
    (df_obs, Dphi_y, Dphi_s, sigma) = run_raking(
        dim=1,
        df_obs=example_1D.df_obs,
        df_margins=[example_1D.df_margin],
        var_names=["var1"],
        cov_mat=False,
    )
    assert np.allclose(
        df_obs["raked_value"].sum(),
        example_1D.df_margin["value_agg_over_var1"].iloc[0],
    ), "The raked values do not sum to the margin."


def test_run_raking_2D(example_2D):
    (df_obs, Dphi_y, Dphi_s, sigma) = run_raking(
        dim=2,
        df_obs=example_2D.df_obs,
        df_margins=[example_2D.df_margins_1, example_2D.df_margins_2],
        var_names=["var1", "var2"],
        cov_mat=False,
    )
    sum_over_var1 = (
        df_obs.groupby(["var2"])
        .agg({"raked_value": "sum"})
        .reset_index()
        .merge(example_2D.df_margins_1, on="var2")
    )
    assert np.allclose(
        sum_over_var1["raked_value"], sum_over_var1["value_agg_over_var1"]
    ), "The sums over the first variable must match the first margins."
    sum_over_var2 = (
        df_obs.groupby(["var1"])
        .agg({"raked_value": "sum"})
        .reset_index()
        .merge(example_2D.df_margins_2, on="var1")
    )
    assert np.allclose(
        sum_over_var2["raked_value"], sum_over_var2["value_agg_over_var2"]
    ), "The sums over the second variable must match the second margins."


def test_run_raking_3D(example_3D):
    (df_obs, Dphi_y, Dphi_s, sigma) = run_raking(
        dim=3,
        df_obs=example_3D.df_obs,
        df_margins=[
            example_3D.df_margins_1,
            example_3D.df_margins_2,
            example_3D.df_margins_3,
        ],
        var_names=["var1", "var2", "var3"],
        cov_mat=False,
    )
    sum_over_var1 = (
        df_obs.groupby(["var2", "var3"])
        .agg({"raked_value": "sum"})
        .reset_index()
        .merge(example_3D.df_margins_1, on=["var2", "var3"])
    )
    assert np.allclose(
        sum_over_var1["raked_value"], sum_over_var1["value_agg_over_var1"]
    ), "The sums over the first variable must match the first margins."
    sum_over_var2 = (
        df_obs.groupby(["var1", "var3"])
        .agg({"raked_value": "sum"})
        .reset_index()
        .merge(example_3D.df_margins_2, on=["var1", "var3"])
    )
    assert np.allclose(
        sum_over_var2["raked_value"], sum_over_var2["value_agg_over_var2"]
    ), "The sums over the second variable must match the second margins."
    sum_over_var3 = (
        df_obs.groupby(["var1", "var2"])
        .agg({"raked_value": "sum"})
        .reset_index()
        .merge(example_3D.df_margins_3, on=["var1", "var2"])
    )
    assert np.allclose(
        sum_over_var3["raked_value"], sum_over_var3["value_agg_over_var3"]
    ), "The sums over the third variable must match the third margins."


def test_run_raking_USHD(example_USHD):
    (df_obs, Dphi_y, Dphi_s, sigma) = run_raking(
        dim="USHD",
        df_obs=example_USHD.df_obs,
        df_margins=[example_USHD.df_margins],
        var_names=None,
        margin_names=["_all", 1, 0],
        cov_mat=False,
    )
    sum_over_cause = (
        df_obs.loc[df_obs.cause != "_all"]
        .groupby(["race", "county"], observed=True)
        .agg({"raked_value": "sum"})
        .reset_index()
        .merge(df_obs.loc[df_obs.cause == "_all"], on=["race", "county"])
    )
    assert np.allclose(
        sum_over_cause["raked_value_x"],
        sum_over_cause["raked_value_y"],
        atol=1.0e-4,
    ), "The sums over the cause must match the all causes deaths."
    sum_over_race = (
        df_obs.loc[df_obs.race != 1]
        .groupby(["cause", "county"], observed=True)
        .agg({"raked_value": "sum"})
        .reset_index()
        .merge(df_obs.loc[df_obs.race == 1], on=["cause", "county"])
    )
    assert np.allclose(
        sum_over_race["raked_value_x"],
        sum_over_race["raked_value_y"],
        atol=1.0e-4,
    ), "The sums over the race must match the all races deaths."
    sum_over_race_county = (
        df_obs.loc[df_obs.race != 1]
        .groupby(["cause"], observed=True)
        .agg({"raked_value": "sum"})
        .reset_index()
        .merge(example_USHD.df_margins, on=["cause"])
    )
    assert np.allclose(
        sum_over_race_county["raked_value"],
        sum_over_race_county["value_agg_over_race_county"],
        atol=1.0e-5,
    ), "The sums over race and county must match the GBD values."


def test_run_raking_USHD_lower(example_USHD_lower):
    (df_obs, Dphi_y, Dphi_s, sigma) = run_raking(
        dim="USHD_lower",
        df_obs=example_USHD_lower.df_obs,
        df_margins=[
            example_USHD_lower.df_margins_cause,
            example_USHD_lower.df_margins_county,
            example_USHD_lower.df_margins_all_causes,
        ],
        var_names=None,
        margin_names=["_inj", 1, 0],
        cov_mat=False,
    )
    sum_over_cause = (
        df_obs.loc[df_obs.race != 1]
        .groupby(["race", "county"], observed=True)
        .agg({"raked_value": "sum"})
        .reset_index()
        .merge(example_USHD_lower.df_margins_all_causes, on=["race", "county"])
    )
    assert np.allclose(
        sum_over_cause["raked_value"],
        sum_over_cause["value_agg_over_cause"],
        atol=1.0e-4,
    ), "The sums over the cause must match the all causes deaths."
    sum_over_cause_race = (
        df_obs.loc[df_obs.race == 1]
        .groupby(["county"], observed=True)
        .agg({"raked_value": "sum"})
        .reset_index()
        .merge(example_USHD_lower.df_margins_county, on=["county"])
    )
    assert np.allclose(
        sum_over_cause_race["raked_value"],
        sum_over_cause_race["value_agg_over_cause_race"],
        atol=1.0e-4,
    ), (
        "The sums over the cause and race must match the all causes all races deaths."
    )
    sum_over_race = (
        df_obs.loc[df_obs.race != 1]
        .groupby(["cause", "county"], observed=True)
        .agg({"raked_value": "sum"})
        .reset_index()
        .merge(df_obs.loc[df_obs.race == 1], on=["cause", "county"])
    )
    assert np.allclose(
        sum_over_race["raked_value_x"],
        sum_over_race["raked_value_y"],
        atol=1.0e-4,
    ), "The sums over the race must match the all races deaths."
    sum_over_race_county = (
        df_obs.loc[df_obs.race != 1]
        .groupby(["cause"], observed=True)
        .agg({"raked_value": "sum"})
        .reset_index()
        .merge(example_USHD_lower.df_margins_cause, on=["cause"])
    )
    assert np.allclose(
        sum_over_race_county["raked_value"],
        sum_over_race_county["value_agg_over_race_county"],
        atol=1.0e-5,
    ), "The sums over race and county must match the GBD values."


def test_run_raking_1D_draws(example_1D_draws):
    (df_obs, Dphi_y, Dphi_s, sigma) = run_raking(
        dim=1,
        df_obs=example_1D_draws.df_obs,
        df_margins=[example_1D_draws.df_margin],
        var_names=["var1"],
        draws="draws",
        cov_mat=True,
    )
    assert np.allclose(
        df_obs["raked_value"].sum(),
        example_1D_draws.df_margin["value_agg_over_var1"].mean(),
    ), "The raked values do not sum to the margin."


def test_run_raking_2D_draws(example_2D_draws):
    (df_obs, Dphi_y, Dphi_s, sigma) = run_raking(
        dim=2,
        df_obs=example_2D_draws.df_obs,
        df_margins=[
            example_2D_draws.df_margins_1,
            example_2D_draws.df_margins_2,
        ],
        var_names=["var1", "var2"],
        draws="draws",
        cov_mat=True,
    )
    sum_over_var1 = (
        df_obs.groupby(["var2"])
        .agg({"raked_value": "sum"})
        .reset_index()
        .merge(
            example_2D_draws.df_margins_1.groupby(["var2"])
            .agg({"value_agg_over_var1": "mean"})
            .reset_index(),
            on="var2",
        )
    )
    assert np.allclose(
        sum_over_var1["raked_value"], sum_over_var1["value_agg_over_var1"]
    ), "The sums over the first variable must match the first margins."
    sum_over_var2 = (
        df_obs.groupby(["var1"])
        .agg({"raked_value": "sum"})
        .reset_index()
        .merge(
            example_2D_draws.df_margins_2.groupby(["var1"])
            .agg({"value_agg_over_var2": "mean"})
            .reset_index(),
            on="var1",
        )
    )
    assert np.allclose(
        sum_over_var2["raked_value"], sum_over_var2["value_agg_over_var2"]
    ), "The sums over the second variable must match the second margins."


def test_run_raking_3D_draws(example_3D_draws):
    (df_obs, Dphi_y, Dphi_s, sigma) = run_raking(
        dim=3,
        df_obs=example_3D_draws.df_obs,
        df_margins=[
            example_3D_draws.df_margins_1,
            example_3D_draws.df_margins_2,
            example_3D_draws.df_margins_3,
        ],
        var_names=["var1", "var2", "var3"],
        draws="draws",
        cov_mat=True,
    )
    sum_over_var1 = (
        df_obs.groupby(["var2", "var3"])
        .agg({"raked_value": "sum"})
        .reset_index()
        .merge(
            example_3D_draws.df_margins_1.groupby(["var2", "var3"])
            .agg({"value_agg_over_var1": "mean"})
            .reset_index(),
            on=["var2", "var3"],
        )
    )
    assert np.allclose(
        sum_over_var1["raked_value"], sum_over_var1["value_agg_over_var1"]
    ), "The sums over the first variable must match the first margins."
    sum_over_var2 = (
        df_obs.groupby(["var1", "var3"])
        .agg({"raked_value": "sum"})
        .reset_index()
        .merge(
            example_3D_draws.df_margins_2.groupby(["var1", "var3"])
            .agg({"value_agg_over_var2": "mean"})
            .reset_index(),
            on=["var1", "var3"],
        )
    )
    assert np.allclose(
        sum_over_var2["raked_value"], sum_over_var2["value_agg_over_var2"]
    ), "The sums over the second variable must match the second margins."
    sum_over_var3 = (
        df_obs.groupby(["var1", "var2"])
        .agg({"raked_value": "sum"})
        .reset_index()
        .merge(
            example_3D_draws.df_margins_3.groupby(["var1", "var2"])
            .agg({"value_agg_over_var3": "mean"})
            .reset_index(),
            on=["var1", "var2"],
        )
    )
    assert np.allclose(
        sum_over_var3["raked_value"], sum_over_var3["value_agg_over_var3"]
    ), "The sums over the third variable must match the third margins."


def test_run_raking_USHD_draws(example_USHD_draws):
    (df_obs, Dphi_y, Dphi_s, sigma) = run_raking(
        dim="USHD",
        df_obs=example_USHD_draws.df_obs,
        df_margins=[example_USHD_draws.df_margins],
        var_names=None,
        margin_names=["_all", 1, 0],
        cov_mat=True,
    )
    sum_over_cause = (
        df_obs.loc[df_obs.cause != "_all"]
        .groupby(["race", "county"], observed=True)
        .agg({"raked_value": "sum"})
        .reset_index()
        .merge(df_obs.loc[df_obs.cause == "_all"], on=["race", "county"])
    )
    assert np.allclose(
        sum_over_cause["raked_value_x"],
        sum_over_cause["raked_value_y"],
        atol=1.0e-4,
    ), "The sums over the cause must match the all causes deaths."
    sum_over_race = (
        df_obs.loc[df_obs.race != 1]
        .groupby(["cause", "county"], observed=True)
        .agg({"raked_value": "sum"})
        .reset_index()
        .merge(df_obs.loc[df_obs.race == 1], on=["cause", "county"])
    )
    assert np.allclose(
        sum_over_race["raked_value_x"],
        sum_over_race["raked_value_y"],
        atol=1.0e-4,
    ), "The sums over the race must match the all races deaths."
    sum_over_race_county = (
        df_obs.loc[df_obs.race != 1]
        .groupby(["cause"], observed=True)
        .agg({"raked_value": "sum"})
        .reset_index()
        .merge(
            example_USHD_draws.df_margins.groupby(["cause"])
            .agg({"value_agg_over_race_county": "mean"})
            .reset_index(),
            on=["cause"],
        )
    )
    assert np.allclose(
        sum_over_race_county["raked_value"],
        sum_over_race_county["value_agg_over_race_county"],
        atol=1.0e-5,
    ), "The sums over race and county must match the GBD values."


def test_run_raking_1D_weights(example_1D_draws):
    df_obs = example_1D_draws.df_obs
    df_obs = (
        df_obs.groupby(["var1"]).agg({"value": ["mean", "std"]}).reset_index()
    )
    df_obs.columns = [" ".join(col).strip() for col in df_obs.columns.values]
    df_obs.rename(
        columns={"value mean": "value", "value std": "weight"}, inplace=True
    )
    df_obs["lower"] = 1.8
    df_obs["upper"] = 3.2
    df_margin = example_1D_draws.df_margin
    df_margin = df_margin[["value_agg_over_var1"]].mean().to_frame().transpose()
    (df_obs, Dphi_y, Dphi_s, sigma) = run_raking(
        dim=1,
        df_obs=df_obs,
        df_margins=[df_margin],
        var_names=["var1"],
        cov_mat=False,
        method="logit",
        weights="weight",
        lower="lower",
        upper="upper",
    )
    assert np.allclose(
        df_obs["raked_value"].sum(),
        df_margin["value_agg_over_var1"].iloc[0],
    ), "The raked values do not sum to the margin."


def test_run_raking_2D_weights(example_2D_draws):
    df_obs = example_2D_draws.df_obs
    df_obs = (
        df_obs.groupby(["var1", "var2"])
        .agg({"value": ["mean", "std"]})
        .reset_index()
    )
    df_obs.columns = [" ".join(col).strip() for col in df_obs.columns.values]
    df_obs.rename(
        columns={"value mean": "value", "value std": "weight"}, inplace=True
    )
    df_obs["lower"] = 1.8
    df_obs["upper"] = 3.2
    df_margins_1 = example_2D_draws.df_margins_1
    df_margins_1 = (
        df_margins_1.groupby(["var2"])
        .agg({"value_agg_over_var1": "mean"})
        .reset_index()
    )
    df_margins_2 = example_2D_draws.df_margins_2
    df_margins_2 = (
        df_margins_2.groupby(["var1"])
        .agg({"value_agg_over_var2": "mean"})
        .reset_index()
    )
    (df_obs, Dphi_y, Dphi_s, sigma) = run_raking(
        dim=2,
        df_obs=df_obs,
        df_margins=[
            df_margins_1,
            df_margins_2,
        ],
        var_names=["var1", "var2"],
        cov_mat=False,
        method="logit",
        weights="weight",
        lower="lower",
        upper="upper",
    )
    sum_over_var1 = (
        df_obs.groupby(["var2"])
        .agg({"raked_value": "sum"})
        .reset_index()
        .merge(df_margins_1, on="var2")
    )
    assert np.allclose(
        sum_over_var1["raked_value"], sum_over_var1["value_agg_over_var1"]
    ), "The sums over the first variable must match the first margins."
    sum_over_var2 = (
        df_obs.groupby(["var1"])
        .agg({"raked_value": "sum"})
        .reset_index()
        .merge(df_margins_2, on="var1")
    )
    assert np.allclose(
        sum_over_var2["raked_value"], sum_over_var2["value_agg_over_var2"]
    ), "The sums over the second variable must match the second margins."


def test_run_raking_3D_weights(example_3D_draws):
    df_obs = example_3D_draws.df_obs
    df_obs = (
        df_obs.groupby(["var1", "var2", "var3"])
        .agg({"value": ["mean", "std"]})
        .reset_index()
    )
    df_obs.columns = [" ".join(col).strip() for col in df_obs.columns.values]
    df_obs.rename(
        columns={"value mean": "value", "value std": "weight"}, inplace=True
    )
    df_obs["lower"] = 1.8
    df_obs["upper"] = 3.2
    df_margins_1 = example_3D_draws.df_margins_1
    df_margins_1 = (
        df_margins_1.groupby(["var2", "var3"])
        .agg({"value_agg_over_var1": "mean"})
        .reset_index()
    )
    df_margins_2 = example_3D_draws.df_margins_2
    df_margins_2 = (
        df_margins_2.groupby(["var1", "var3"])
        .agg({"value_agg_over_var2": "mean"})
        .reset_index()
    )
    df_margins_3 = example_3D_draws.df_margins_3
    df_margins_3 = (
        df_margins_3.groupby(["var1", "var2"])
        .agg({"value_agg_over_var3": "mean"})
        .reset_index()
    )
    (df_obs, Dphi_y, Dphi_s, sigma) = run_raking(
        dim=3,
        df_obs=df_obs,
        df_margins=[
            df_margins_1,
            df_margins_2,
            df_margins_3,
        ],
        var_names=["var1", "var2", "var3"],
        cov_mat=False,
        method="logit",
        weights="weight",
        lower="lower",
        upper="upper",
    )
    sum_over_var1 = (
        df_obs.groupby(["var2", "var3"])
        .agg({"raked_value": "sum"})
        .reset_index()
        .merge(df_margins_1, on=["var2", "var3"])
    )
    assert np.allclose(
        sum_over_var1["raked_value"], sum_over_var1["value_agg_over_var1"]
    ), "The sums over the first variable must match the first margins."
    sum_over_var2 = (
        df_obs.groupby(["var1", "var3"])
        .agg({"raked_value": "sum"})
        .reset_index()
        .merge(df_margins_2, on=["var1", "var3"])
    )
    assert np.allclose(
        sum_over_var2["raked_value"], sum_over_var2["value_agg_over_var2"]
    ), "The sums over the second variable must match the second margins."
    sum_over_var3 = (
        df_obs.groupby(["var1", "var2"])
        .agg({"raked_value": "sum"})
        .reset_index()
        .merge(df_margins_3, on=["var1", "var2"])
    )
    assert np.allclose(
        sum_over_var3["raked_value"], sum_over_var3["value_agg_over_var3"]
    ), "The sums over the third variable must match the third margins."


def test_run_raking_USHD_weights(example_USHD_draws):
    df_obs = example_USHD_draws.df_obs
    df_obs = (
        df_obs.groupby(["cause", "race", "county", "upper"])
        .agg({"value": ["mean", "std"]})
        .reset_index()
    )
    df_obs.columns = [" ".join(col).strip() for col in df_obs.columns.values]
    df_obs.rename(
        columns={"value mean": "value", "value std": "weight"}, inplace=True
    )
    df_obs["lower"] = 0.0
    df_margins = example_USHD_draws.df_margins
    df_margins = (
        df_margins.groupby(["cause"])
        .agg({"value_agg_over_race_county": "mean"})
        .reset_index()
    )
    (df_obs, Dphi_y, Dphi_s, sigma) = run_raking(
        dim="USHD",
        df_obs=df_obs,
        df_margins=[df_margins],
        var_names=None,
        margin_names=["_all", 1, 0],
        cov_mat=False,
        method="logit",
        weights="weight",
        lower="lower",
        upper="upper",
    )
    sum_over_cause = (
        df_obs.loc[df_obs.cause != "_all"]
        .groupby(["race", "county"], observed=True)
        .agg({"raked_value": "sum"})
        .reset_index()
        .merge(df_obs.loc[df_obs.cause == "_all"], on=["race", "county"])
    )
    assert np.allclose(
        sum_over_cause["raked_value_x"],
        sum_over_cause["raked_value_y"],
        atol=1.0e-4,
    ), "The sums over the cause must match the all causes deaths."
    sum_over_race = (
        df_obs.loc[df_obs.race != 1]
        .groupby(["cause", "county"], observed=True)
        .agg({"raked_value": "sum"})
        .reset_index()
        .merge(df_obs.loc[df_obs.race == 1], on=["cause", "county"])
    )
    assert np.allclose(
        sum_over_race["raked_value_x"],
        sum_over_race["raked_value_y"],
        atol=1.0e-4,
    ), "The sums over the race must match the all races deaths."
    sum_over_race_county = (
        df_obs.loc[df_obs.race != 1]
        .groupby(["cause"], observed=True)
        .agg({"raked_value": "sum"})
        .reset_index()
        .merge(df_margins, on=["cause"])
    )
    assert np.allclose(
        sum_over_race_county["raked_value"],
        sum_over_race_county["value_agg_over_race_county"],
        atol=1.0e-5,
    ), "The sums over race and county must match the GBD values."


def test_run_raking_USHD_lower_weights(example_USHD_lower_draws):
    df_obs = example_USHD_lower_draws.df_obs
    df_obs = (
        df_obs.groupby(["cause", "race", "county", "upper"])
        .agg({"value": ["mean", "std"]})
        .reset_index()
    )
    df_obs.columns = [" ".join(col).strip() for col in df_obs.columns.values]
    df_obs.rename(
        columns={"value mean": "value", "value std": "weight"}, inplace=True
    )
    df_obs["lower"] = 0.0
    df_margins_cause = example_USHD_lower_draws.df_margins_cause
    df_margins_cause = (
        df_margins_cause.groupby(["cause"])
        .agg({"value_agg_over_race_county": "mean"})
        .reset_index()
    )
    df_margins_county = example_USHD_lower_draws.df_margins_county
    df_margins_county = (
        df_margins_county.groupby(["county"])
        .agg({"value_agg_over_cause_race": "mean"})
        .reset_index()
    )
    df_margins_all_causes = example_USHD_lower_draws.df_margins_all_causes
    df_margins_all_causes = (
        df_margins_all_causes.groupby(["race", "county"])
        .agg({"value_agg_over_cause": "mean"})
        .reset_index()
    )
    (df_obs, Dphi_y, Dphi_s, sigma) = run_raking(
        dim="USHD_lower",
        df_obs=df_obs,
        df_margins=[
            df_margins_cause,
            df_margins_county,
            df_margins_all_causes,
        ],
        var_names=None,
        margin_names=["_inj", 1, 0],
        cov_mat=False,
        method="logit",
        weights="weight",
        lower="lower",
        upper="upper",
    )
    sum_over_cause = (
        df_obs.loc[df_obs.race != 1]
        .groupby(["race", "county"], observed=True)
        .agg({"raked_value": "sum"})
        .reset_index()
        .merge(df_margins_all_causes, on=["race", "county"])
    )
    assert np.allclose(
        sum_over_cause["raked_value"],
        sum_over_cause["value_agg_over_cause"],
        atol=1.0e-4,
    ), "The sums over the cause must match the all causes deaths."
    sum_over_cause_race = (
        df_obs.loc[df_obs.race == 1]
        .groupby(["county"], observed=True)
        .agg({"raked_value": "sum"})
        .reset_index()
        .merge(df_margins_county, on=["county"])
    )
    assert np.allclose(
        sum_over_cause_race["raked_value"],
        sum_over_cause_race["value_agg_over_cause_race"],
        atol=1.0e-4,
    ), (
        "The sums over the cause and race must match the all causes all races deaths."
    )
    sum_over_race = (
        df_obs.loc[df_obs.race != 1]
        .groupby(["cause", "county"], observed=True)
        .agg({"raked_value": "sum"})
        .reset_index()
        .merge(df_obs.loc[df_obs.race == 1], on=["cause", "county"])
    )
    assert np.allclose(
        sum_over_race["raked_value_x"],
        sum_over_race["raked_value_y"],
        atol=1.0e-4,
    ), "The sums over the race must match the all races deaths."
    sum_over_race_county = (
        df_obs.loc[df_obs.race != 1]
        .groupby(["cause"], observed=True)
        .agg({"raked_value": "sum"})
        .reset_index()
        .merge(df_margins_cause, on=["cause"])
    )
    assert np.allclose(
        sum_over_race_county["raked_value"],
        sum_over_race_county["value_agg_over_race_county"],
        atol=1.0e-5,
    ), "The sums over race and county must match the GBD values."



import pytest
import numpy as np
import pandas as pd
from raking.compute_constraints import (
    constraints_1D,
    constraints_2D,
    constraints_3D,
    constraints_USHD,
    constraints_USHD_lower,
)
from raking.formatting_methods import (
    format_data_1D,
    format_data_2D,
    format_data_3D,
    format_data_USHD,
    format_data_USHD_lower,
)


def test_format_data_1D():
    # Generate balanced vector
    I = 3
    rng = np.random.default_rng(0)
    y = rng.uniform(low=2.0, high=3.0, size=I)
    s = np.sum(y)
    # Generate the data frames
    var1 = np.arange(0, I)
    df_obs = pd.DataFrame({"value": y, "var1": var1})
    df_margins = pd.DataFrame({"value_agg_over_var1": [s]})
    # Get the formatted data
    (y, s, I, q, l, h) = format_data_1D(df_obs, df_margins, "var1")
    # Generate the constraints
    (A, s) = constraints_1D(s, I)
    # Verify that the constraint A beta = s is respected
    assert np.allclose(np.matmul(A, y), s), (
        "For the format_data_1D function, the constraint A y = s is not respected."
    )


def test_format_data_2D():
    # Generate balanced matrix
    I = 3
    J = 5
    rng = np.random.default_rng(0)
    y = rng.uniform(low=2.0, high=3.0, size=(I, J))
    s1 = np.sum(y, axis=0)
    s2 = np.sum(y, axis=1)
    y = y.flatten(order="F")
    # Generate the data frames
    var1 = np.tile(np.arange(0, I), J)
    var2 = np.repeat(np.arange(0, J), I)
    df_obs = pd.DataFrame({"value": y, "var1": var1, "var2": var2})
    df_margins_1 = pd.DataFrame(
        {"var2": np.arange(0, J), "value_agg_over_var1": s1}
    )
    df_margins_2 = pd.DataFrame(
        {"var1": np.arange(0, I), "value_agg_over_var2": s2}
    )
    # Get the formatted data
    (y, s1, s2, I, J, q, l, h) = format_data_2D(
        df_obs, df_margins_1, df_margins_2, ["var1", "var2"]
    )
    # Generate the constraints
    (A, s) = constraints_2D(s1, s2, I, J)
    # Verify that the constraint A beta = s is respected
    assert np.allclose(np.matmul(A, y), s), (
        "For the format_data_2D function, the constraint A y = s is not respected."
    )


def test_format_data_3D():
    # Generate balanced array
    I = 3
    J = 4
    K = 5
    rng = np.random.default_rng(0)
    y = rng.uniform(low=2.0, high=3.0, size=(I, J, K))
    s1 = np.sum(y, axis=0)
    s2 = np.sum(y, axis=1)
    s3 = np.sum(y, axis=2)
    y = y.flatten(order="F")
    s1 = s1.flatten(order="F")
    s2 = s2.flatten(order="F")
    s3 = s3.flatten(order="F")
    # Generate the observation data frame
    var1 = np.tile(np.arange(0, I), J * K)
    var2 = np.tile(np.repeat(np.arange(0, J), I), K)
    var3 = np.repeat(np.arange(0, K), I * J)
    df_obs = pd.DataFrame(
        {"value": y, "var1": var1, "var2": var2, "var3": var3}
    )
    # Generate the first margins data frame
    var2 = np.tile(np.arange(0, J), K)
    var3 = np.repeat(np.arange(0, K), J)
    df_margins_1 = pd.DataFrame(
        {"var2": var2, "var3": var3, "value_agg_over_var1": s1}
    )
    # Generate the second margins data frame
    var1 = np.tile(np.arange(0, I), K)
    var3 = np.repeat(np.arange(0, K), I)
    df_margins_2 = pd.DataFrame(
        {"var1": var1, "var3": var3, "value_agg_over_var2": s2}
    )
    # Generate the third margins data frame
    var1 = np.tile(np.arange(0, I), J)
    var2 = np.repeat(np.arange(0, J), I)
    df_margins_3 = pd.DataFrame(
        {"var1": var1, "var2": var2, "value_agg_over_var3": s3}
    )
    # Get the formatted data
    (y, s1, s2, s3, I, J, K, q, l, h) = format_data_3D(
        df_obs,
        df_margins_1,
        df_margins_2,
        df_margins_3,
        ["var1", "var2", "var3"],
    )
    # Generate the constraints
    (A, s) = constraints_3D(s1, s2, s3, I, J, K)
    # Verify that the constraint A beta = s is respected
    assert np.allclose(np.matmul(A, y), s), (
        "For the format_data_3D function, the constraint A y = s is not respected."
    )


def test_format_data_USHD():
    # Generate balanced array
    I = 3
    J = 4
    K = 5
    rng = np.random.default_rng(0)
    y_ijk = rng.uniform(low=2.0, high=3.0, size=(I, J, K))
    y_00k = np.sum(y_ijk, axis=(0, 1))
    y_i0k = np.sum(y_ijk, axis=1)
    y_0jk = np.sum(y_ijk, axis=0)
    y1 = np.concatenate(
        (y_00k.reshape((1, 1, K)), y_i0k.reshape(I, 1, K)), axis=0
    )
    y2 = np.concatenate((y_0jk.reshape((1, J, K)), y_ijk), axis=0)
    y = np.concatenate((y1, y2), axis=1)
    y_i = np.sum(y_ijk, axis=(1, 2))
    y_0 = np.sum(y_i)
    s_cause = np.array([y_0] + y_i.tolist())
    y = y.flatten(order="F")
    # Generate the observation data frame
    cause = np.tile(np.arange(0, I + 1), (J + 1) * K)
    race = np.tile(np.repeat(np.arange(0, J + 1), I + 1), K)
    county = np.repeat(np.arange(1, K + 1), (I + 1) * (J + 1))
    df_obs = pd.DataFrame(
        {"value": y, "cause": cause, "race": race, "county": county}
    )
    # Generate the margins data frame
    cause = np.arange(0, I + 1)
    df_margins = pd.DataFrame(
        {"cause": cause, "value_agg_over_race_county": s_cause}
    )
    # Get the formatted data
    (y, s, I, J, K, q, l, h) = format_data_USHD(df_obs, df_margins, [0, 0, 0])
    # Generate the constraints
    (A, s) = constraints_USHD(s, I, J, K)
    # Verify that the constraint A beta = s is respected
    assert np.allclose(np.matmul(A, y), s), (
        "For the format_data_USHD function, the constraint A y = s is not respected."
    )


def test_format_data_USHD_lower():
    # Generate balanced array
    I = 3
    J = 4
    K = 5
    rng = np.random.default_rng(0)
    y_ijk = rng.uniform(low=2.0, high=3.0, size=(I, J, K))
    s_cause = np.sum(y_ijk, axis=(1, 2))
    s_county = np.sum(y_ijk, axis=(0, 1))
    s_all_causes = np.sum(y_ijk, axis=0)
    y_i0k = np.sum(y_ijk, axis=1)
    y = np.concatenate((y_i0k.reshape((I, 1, K)), y_ijk), axis=1)
    y = y.flatten("F")
    # Generate the observation data frame
    cause = np.tile(np.arange(1, I + 1), (J + 1) * K)
    race = np.tile(np.repeat(np.arange(0, J + 1), I), K)
    county = np.repeat(np.arange(1, K + 1), I * (J + 1))
    df_obs = pd.DataFrame(
        {"value": y, "cause": cause, "race": race, "county": county}
    )
    # Generate the GBD margins data frame
    cause = np.arange(1, I + 1)
    df_margins_cause = pd.DataFrame(
        {"cause": cause, "value_agg_over_race_county": s_cause}
    )
    # Generate the county margins data frame
    county = np.arange(1, K + 1)
    df_margins_county = pd.DataFrame(
        {"county": county, "value_agg_over_cause_race": s_county}
    )
    # Generate the all causes margins data frame
    race = np.tile(np.arange(1, J + 1), K)
    county = np.repeat(np.arange(1, K + 1), J)
    df_margins_all_causes = pd.DataFrame(
        {
            "race": race,
            "county": county,
            "value_agg_over_cause": s_all_causes.flatten("F"),
        }
    )
    # Get the formatted data
    (y, s_cause, s_county, s_all_causes, I, J, K, q, l, h) = (
        format_data_USHD_lower(
            df_obs,
            df_margins_cause,
            df_margins_county,
            df_margins_all_causes,
            [0, 0, 0],
        )
    )
    # Generate the constraints
    (A, s) = constraints_USHD_lower(s_cause, s_county, s_all_causes, I, J, K)
    # Verify that the constraint A beta = s is respected
    assert np.allclose(np.matmul(A, y), s), (
        "For the format_data_USHD_lower function, the constraint A y = s is not respected."
    )



"""Module with methods to solve the raking problem"""

import numpy as np
from scipy.sparse.linalg import cg


def raking_chi2(
    y: np.ndarray,
    A: np.ndarray,
    s: np.ndarray,
    q: np.ndarray = None,
) -> tuple[np.ndarray, np.ndarray]:
    """Raking using the chi2 distance f(beta, y) = (beta - y)^2 / 2y.

    This will solve the problem:
        min_beta 1/q f(beta, y) s.t. A beta = s

    Parameters
    ----------
    y : np.ndarray
        Vector of observations
    A : np.ndarray
        Constraints matrix (output of a function from the compute_constraints module)
    s : np.ndarray
        Margin vector (output of a function from the compute_constraints module)
    q : np.ndarray
        Vector of weights (default to all 1)

    Returns
    -------
    beta : np.ndarray
        Vector of raked values
    lambda_k : np.ndarray
        Dual (needed for the uncertainty computation)
    """
    assert isinstance(y, np.ndarray), (
        "The vector of observations should be a Numpy array."
    )
    assert len(y.shape) == 1, (
        "The vector of observations should be a 1D Numpy array."
    )
    if q is not None:
        assert isinstance(q, np.ndarray), (
            "The vector of weights should be a Numpy array."
        )
        assert len(y.shape) == 1, (
            "The vector of weights should be a 1D Numpy array."
        )
        assert len(y) == len(q), (
            "Observations and weights vectors should have the same length."
        )
    assert isinstance(A, np.ndarray), (
        "The constraint matrix should be a Numpy array."
    )
    assert len(A.shape) == 2, (
        "The constraints matrix should be a 2D Numpy array."
    )
    assert isinstance(s, np.ndarray), (
        "The margins vector should be a Numpy array."
    )
    assert len(s.shape) == 1, "The margins vector should be a 1D Numpy array."
    assert np.shape(A)[0] == len(s), (
        "The number of linear constraints should be equal to the number of margins."
    )
    assert np.shape(A)[1] == len(y), (
        "The number of coefficients for the linear constraints should be equal to the number of observations."
    )

    if q is None:
        q = np.ones(len(y))
    s_hat = np.matmul(A, y)
    Phi = np.matmul(A, np.transpose(A * y * q))
    lambda_k = cg(Phi, s_hat - s)[0]
    beta = y * (1 - q * np.matmul(np.transpose(A), lambda_k))
    return (beta, lambda_k)


def raking_entropic(
    y: np.ndarray,
    A: np.ndarray,
    s: np.ndarray,
    q: np.ndarray = None,
    gamma0: float = 1.0,
    max_iter: int = 500,
) -> tuple[np.ndarray, np.ndarray, int]:
    """Raking using the entropic distance f(beta, y) = beta log(beta/y) + y - beta.

    This will solve the problem:
        min_beta 1/q f(beta, y) s.t. A beta = s

    Parameters
    ----------
    y : np.ndarray
        Vector of observations
    A : np.ndarray
        Constraints matrix (output of a function from the compute_constraints module)
    s : np.ndarray
        Margin vector (output of a function from the compute_constraints module)
    q : np.ndarray
        Vector of weights (default to all 1)
    gamma0 : float
        Initial value for line search
    max_iter : int
        Number of iterations for Newton's root finding method

    Returns
    -------
    beta : np.ndarray
        Vector of reaked values
    lambda_k : np.ndarray
        Dual (needed for th uncertainty computation)
    iters_eps : int
        Number of iterations until convergence
    """
    assert isinstance(y, np.ndarray), (
        "The vector of observations should be a Numpy array."
    )
    assert len(y.shape) == 1, (
        "The vector of observations should be a 1D Numpy array."
    )
    if q is not None:
        assert isinstance(q, np.ndarray), (
            "The vector of weights should be a Numpy array."
        )
        assert len(q.shape) == 1, (
            "The vector of weights should be a 1D Numpy array."
        )
        assert len(y) == len(q), (
            "Observations and weights vectors should have the same length."
        )
    assert isinstance(A, np.ndarray), (
        "The constraint matrix should be a Numpy array."
    )
    assert len(A.shape) == 2, (
        "The constraints matrix should be a 2D Numpy array."
    )
    assert isinstance(s, np.ndarray), (
        "The margins vector should be a Numpy array."
    )
    assert len(s.shape) == 1, "The margins vector should be a 1D Numpy array."
    assert np.shape(A)[0] == len(s), (
        "The number of linear constraints should be equal to the number of margins."
    )
    assert np.shape(A)[1] == len(y), (
        "The number of coefficients for the linear constraints should be equal to the number of observations."
    )

    if q is None:
        q = np.ones(len(y))
    s_hat = np.matmul(A, y)
    lambda_k = np.zeros(A.shape[0])
    beta = np.copy(y)
    epsilon = 1.0
    iter_eps = 0
    while (epsilon > 1.0e-10) & (iter_eps < max_iter):
        Phi = np.matmul(
            A, y * (1.0 - np.exp(-q * np.matmul(np.transpose(A), lambda_k)))
        )
        D = y * q * np.exp(-q * np.matmul(np.transpose(A), lambda_k))
        J = np.matmul(A * D, np.transpose(A))
        delta_lambda = cg(J, Phi - s_hat + s)[0]
        gamma = gamma0
        iter_gam = 0
        lambda_k = lambda_k - gamma * delta_lambda
        beta = y * np.exp(-q * np.matmul(np.transpose(A), lambda_k))
        if iter_eps > 0:
            while (np.mean(np.abs(s - np.matmul(A, beta))) > epsilon) & (
                iter_gam < max_iter
            ):
                gamma = gamma / 2.0
                iter_gam = iter_gam + 1
                lambda_k = lambda_k - gamma * delta_lambda
                beta = y * np.exp(-q * np.matmul(np.transpose(A), lambda_k))
        epsilon = np.mean(np.abs(s - np.matmul(A, beta)))
        iter_eps = iter_eps + 1
    return (beta, lambda_k, iter_eps)


def raking_general(
    y: np.ndarray,
    A: np.ndarray,
    s: np.ndarray,
    alpha: float = 1,
    q: np.ndarray = None,
    gamma0: float = 1.0,
    max_iter: int = 500,
) -> tuple[np.ndarray, np.ndarray, int]:
    """Raking using the general distance f(beta, y) = 1/alpha (y/alpha+1 (beta/y)^alpha+1 - beta + c(y)).

    This will solve the problem:
        min_beta 1/q f(beta, y) s.t. A beta = s

    Parameters
    ----------
    y : np.ndarray
        Vector of observations
    A : np.ndarray
        Constraints matrix (output of a function from the compute_constraints module)
    s : np.ndarray
        Margin vector (output of a function from the compute_constraints module)
    alpha : float
        Parameter of the distance function, alpha=1 is the chi2 distance, alpha=0 is the entropic distance
    q : np.ndarray
        Vector of weights (default to all 1)
    gamma0 : float
        Initial value for line search
    max_iter : int
        Number of iterations for Newton's root finding method

    Returns
    -------
    beta : np.ndarray
        Vector of reaked values
    lambda_k : np.ndarray
        Dual (needed for th uncertainty computation)
    iters_eps : int
        Number of iterations until convergence
    """
    assert isinstance(y, np.ndarray), (
        "The vector of observations should be a Numpy array."
    )
    assert len(y.shape) == 1, (
        "The vector of observations should be a 1D Numpy array."
    )
    if q is not None:
        assert isinstance(q, np.ndarray), (
            "The vector of weights should be a Numpy array."
        )
        assert len(y.shape) == 1, (
            "The vector of weights should be a 1D Numpy array."
        )
        assert len(y) == len(q), (
            "Observations and weights vectors should have the same length."
        )
    assert isinstance(A, np.ndarray), (
        "The constraint matrix should be a Numpy array."
    )
    assert len(A.shape) == 2, (
        "The constraints matrix should be a 2D Numpy array."
    )
    assert isinstance(s, np.ndarray), (
        "The margins vector should be a Numpy array."
    )
    assert len(s.shape) == 1, "The margins vector should be a 1D Numpy array."
    assert np.shape(A)[0] == len(s), (
        "The number of linear constraints should be equal to the number of margins."
    )
    assert np.shape(A)[1] == len(y), (
        "The number of coefficients for the linear constraints should be equal to the number of observations."
    )
    assert isinstance(alpha, (int, float)), (
        "The parameter of the distance function should be an integer or a float."
    )

    if q is None:
        q = np.ones(len(y))

    if alpha == 1:
        (beta, lambda_k) = raking_chi2(y, A, s, q)
        return (beta, lambda_k, 0)

    if alpha == 0:
        (beta, lambda_k, iter_eps) = raking_entropic(
            y, A, s, q, gamma0, max_iter
        )
        return (beta, lambda_k, iter_eps)

    s_hat = np.matmul(A, y)
    lambda_k = np.zeros(A.shape[0])
    beta = np.copy(y)
    epsilon = 1.0
    iter_eps = 0
    while (epsilon > 1.0e-10) & (iter_eps < max_iter):
        Phi = np.matmul(
            A,
            y
            * (
                1.0
                - np.power(
                    1 - alpha * q * np.matmul(np.transpose(A), lambda_k),
                    1.0 / alpha,
                )
            ),
        )
        D = (
            y
            * q
            * np.power(
                1.0 - alpha * q * np.matmul(np.transpose(A), lambda_k),
                1.0 / alpha - 1,
            )
        )
        J = np.matmul(A * D, np.transpose(A))
        delta_lambda = cg(J, Phi - s_hat + s)[0]
        gamma = gamma0
        iter_gam = 0
        lambda_k = lambda_k - gamma * delta_lambda
        beta = y * np.power(
            1.0 - alpha * q * np.matmul(np.transpose(A), lambda_k), 1.0 / alpha
        )
        if (alpha > 0.5) or (alpha < -1.0):
            if iter_eps > 0:
                while (
                    (
                        np.any(
                            1
                            - alpha
                            * q
                            * np.matmul(
                                np.transpose(A), lambda_k - gamma * delta_lambda
                            )
                            <= 0.0
                        )
                    )
                    & (np.mean(np.abs(s - np.matmul(A, beta))) > epsilon)
                    & (iter_gam < max_iter)
                ):
                    gamma = gamma / 2.0
                    iter_gam = iter_gam + 1
                    lambda_k = lambda_k - gamma * delta_lambda
                    beta = y * np.power(
                        1.0 - alpha * q * np.matmul(np.transpose(A), lambda_k),
                        1.0 / alpha,
                    )
            else:
                while (
                    np.any(
                        1
                        - alpha
                        * q
                        * np.matmul(
                            np.transpose(A), lambda_k - gamma * delta_lambda
                        )
                        <= 0.0
                    )
                ) & (iter_gam < max_iter):
                    gamma = gamma / 2.0
                    iter_gam = iter_gam + 1
                    lambda_k = lambda_k - gamma * delta_lambda
                    beta = y * np.power(
                        1.0 - alpha * q * np.matmul(np.transpose(A), lambda_k),
                        1.0 / alpha,
                    )
        else:
            if iter_eps > 0:
                while (np.mean(np.abs(s - np.matmul(A, beta))) > epsilon) & (
                    iter_gam < max_iter
                ):
                    gamma = gamma / 2.0
                    iter_gam = iter_gam + 1
                    lambda_k = lambda_k - gamma * delta_lambda
                    beta = y * np.power(
                        1.0 - alpha * q * np.matmul(np.transpose(A), lambda_k),
                        1.0 / alpha,
                    )
        epsilon = np.mean(np.abs(s - np.matmul(A, beta)))
        iter_eps = iter_eps + 1
    return (beta, lambda_k, iter_eps)


def raking_logit(
    y: np.ndarray,
    A: np.ndarray,
    s: np.ndarray,
    l: np.ndarray = None,
    h: np.ndarray = None,
    q: np.ndarray = None,
    gamma0: float = 1.0,
    max_iter: int = 500,
) -> tuple[np.ndarray, np.ndarray, int]:
    """Logit raking ensuring that l < beta < h.

    This will solve the problem:
        min_beta 1/q f(beta, y) s.t. A beta = s

    Parameters
    ----------
    y : np.ndarray
        Vector of observations
    A : np.ndarray
        Constraints matrix (output of a function from the compute_constraints module)
    s : np.ndarray
        Margin vector (output of a function from the compute_constraints module)
    l : np.ndarray
        Lower bounds for the observations
    h : np.ndarray
        Upper bounds for the observations
    q : np.ndarray
        Vector of weights (default to all 1)
    gamma0 : float
        Initial value for line search
    max_iter : int
        Number of iterations for Newton's root finding method

    Returns
    -------
    beta : np.ndarray
        Vector of reaked values
    lambda_k : np.ndarray
        Dual (needed for th uncertainty computation)
    iters_eps : int
        Number of iterations until convergence
    """
    assert isinstance(y, np.ndarray), (
        "The vector of observations should be a Numpy array."
    )
    assert len(y.shape) == 1, (
        "The vector of observations should be a 1D Numpy array."
    )
    if q is not None:
        assert isinstance(q, np.ndarray), (
            "The vector of weights should be a Numpy array."
        )
        assert len(y.shape) == 1, (
            "The vector of weights should be a 1D Numpy array."
        )
        assert len(y) == len(q), (
            "Observations and weights vectors should have the same length."
        )
    assert isinstance(A, np.ndarray), (
        "The constraint matrix should be a Numpy array."
    )
    assert len(A.shape) == 2, (
        "The constraints matrix should be a 2D Numpy array."
    )
    assert isinstance(s, np.ndarray), (
        "The margins vector should be a Numpy array."
    )
    assert len(s.shape) == 1, "The margins vector should be a 1D Numpy array."
    assert np.shape(A)[0] == len(s), (
        "The number of linear constraints should be equal to the number of margins."
    )
    assert np.shape(A)[1] == len(y), (
        "The number of coefficients for the linear constraints should be equal to the number of observations."
    )

    if l is None:
        l = np.zeros(len(y))
    assert isinstance(l, np.ndarray), (
        "The vector of lower bounds should be a Numpy array."
    )
    assert len(l.shape) == 1, (
        "The vector of lower bounds should be a 1D Numpy array."
    )
    assert len(y) == len(l), (
        "Observations and lower bounds vectors should have the same length."
    )
    assert np.all(l >= 0.0), "The lower bounds must be positive."
    assert np.all(l <= y), (
        "The observations must be superior or equal to the corresponding lower bounds."
    )

    if h is None:
        h = np.ones(len(y))
    assert isinstance(h, np.ndarray), (
        "The vector of upper bounds should be a Numpy array."
    )
    assert len(h.shape) == 1, (
        "The vector of upper bounds should be a 1D Numpy array."
    )
    assert len(y) == len(h), (
        "Observations and upper bounds vectors should have the same length."
    )
    assert np.all(h > 0.0), "The upper bounds must be strictly positive."
    assert np.all(h >= y), (
        "The observations must be inferior or equal to the correspondings upper bounds."
    )

    assert np.all(l < h), (
        "The lower bounds must be stricty inferior to the correspondings upper bounds."
    )

    if q is None:
        q = np.ones(len(y))
    lambda_k = np.zeros(A.shape[0])
    beta = np.copy(y)
    epsilon = 1.0
    iter_eps = 0
    while (epsilon > 1.0e-10) & (iter_eps < max_iter):
        Phi = np.matmul(
            A,
            (
                l * (h - y)
                + h
                * (y - l)
                * np.exp(-q * np.matmul(np.transpose(A), lambda_k))
            )
            / (
                (h - y)
                + (y - l) * np.exp(-q * np.matmul(np.transpose(A), lambda_k))
            ),
        )
        D = (
            -q
            * ((y - l) * (h - y) * (h - l))
            / np.square(
                (h - y)
                + (y - l) * np.exp(-q * np.matmul(np.transpose(A), lambda_k))
            )
        )
        J = np.matmul(A * D, np.transpose(A))
        delta_lambda = cg(J, Phi - s)[0]
        gamma = gamma0
        iter_gam = 0
        lambda_k = lambda_k - gamma * delta_lambda
        beta = (
            l * (h - y)
            + h * (y - l) * np.exp(-q * np.matmul(np.transpose(A), lambda_k))
        ) / (
            (h - y)
            + (y - l) * np.exp(-q * np.matmul(np.transpose(A), lambda_k))
        )
        if iter_eps > 0:
            while (np.mean(np.abs(s - np.matmul(A, beta))) > epsilon) & (
                iter_gam < max_iter
            ):
                gamma = gamma / 2.0
                iter_gam = iter_gam + 1
                lambda_k = lambda_k - gamma * delta_lambda
                beta = (
                    l * (h - y)
                    + h
                    * (y - l)
                    * np.exp(-q * np.matmul(np.transpose(A), lambda_k))
                ) / (
                    (h - y)
                    + (y - l)
                    * np.exp(-q * np.matmul(np.transpose(A), lambda_k))
                )
        epsilon = np.mean(np.abs(s - np.matmul(A, beta)))
        iter_eps = iter_eps + 1
    return (beta, lambda_k, iter_eps)
